import math, random, sys


class ElasticHashTable:
    def __init__(self, capacity, delta=0.1):
        if capacity <= 0:
            raise ValueError("Capacity must be positive.")
        if not (0 < delta < 1):
            raise ValueError("delta must be between 0 and 1.")
        self.capacity = capacity
        self.delta = delta
        self.max_inserts = capacity - int(delta * capacity)
        self.num_inserts = 0

        num_levels = max(1, math.floor(math.log2(capacity)))
        sizes = []
        remaining = capacity
        for i in range(num_levels - 1):
            size = max(1, remaining // (2 ** (num_levels - i)))
            sizes.append(size)
            remaining -= size
        sizes.append(remaining)
        self.levels = [[None] * s for s in sizes]
        self.salts = [random.randint(0, sys.maxsize) for _ in range(num_levels)]
        self.occupancies = [0] * num_levels
        self.c = 4

    def _hash(self, key, level):
        return hash((key, self.salts[level])) & 0x7FFFFFFF

    def _quad_probe(self, key, level, j, table_size):
        # Returns the index for the j-th quadratic probe.
        return (self._hash(key, level) + j * j) % table_size

    def __setitem__(self, key, value):
        self.insert(key, value)

    def insert(self, key, value):
        if self.num_inserts >= self.max_inserts:
            raise RuntimeError(
                "Hash table is full (maximum allowed insertions reached)."
            )
        for i, level in enumerate(self.levels):
            size = len(level)
            occ = self.occupancies[i]
            free = size - occ
            load = free / size
            probe_limit = int(
                max(
                    1,
                    self.c
                    * min(
                        math.log2(1 / load) if load > 0 else 0,
                        math.log2(1 / self.delta),
                    ),
                )
            )
            if i < len(self.levels) - 1:
                next_level = self.levels[i + 1]
                next_occ = self.occupancies[i + 1]
                next_free = len(next_level) - next_occ
                load_next = (next_free / len(next_level)) if len(next_level) > 0 else 0
                threshold = 0.25
                if load > (self.delta / 2) and load_next > threshold:
                    for j in range(probe_limit):
                        idx = self._quad_probe(key, i, j, size)
                        if level[idx] is None:
                            level[idx] = (key, value)
                            self.occupancies[i] += 1
                            self.num_inserts += 1
                            return True
                    # Try next level if not successful.
                elif load <= (self.delta / 2):
                    continue
                elif load_next <= threshold:
                    for j in range(size):
                        idx = self._quad_probe(key, i, j, size)
                        if level[idx] is None:
                            level[idx] = (key, value)
                            self.occupancies[i] += 1
                            self.num_inserts += 1
                            return True
            else:
                for j in range(size):
                    idx = self._quad_probe(key, i, j, size)
                    if level[idx] is None:
                        level[idx] = (key, value)
                        self.occupancies[i] += 1
                        self.num_inserts += 1
                        return True
        raise RuntimeError("Insertion failed in all levels; hash table is full.")

    def __getitem__(self, key):
        ret = self.search(key)
        if ret is None:
            raise KeyError(key)
        return ret

    def get(self, key, default=None):
        return self.search(key) or default

    def search(self, key):
        for i, level in enumerate(self.levels):
            size = len(level)
            for j in range(size):
                idx = self._quad_probe(key, i, j, size)
                entry = level[idx]
                if entry is None:
                    break
                if entry[0] == key:
                    return entry[1]
        return None

    def __contains__(self, key):
        return self.search(key) is not None

    def __len__(self):
        return self.num_inserts



import math
import random
import sys


class FunnelHashTable:
    def __init__(self, capacity, delta=0.1):
        if capacity <= 0:
            raise ValueError("Capacity must be positive.")
        if not (0 < delta < 1):
            raise ValueError("delta must be between 0 and 1.")
        self.capacity = capacity
        self.delta = delta
        self.max_inserts = capacity - int(delta * capacity)
        self.num_inserts = 0

        self.alpha = int(math.ceil(4 * math.log2(1 / delta) + 10))
        self.beta = int(math.ceil(2 * math.log2(1 / delta)))

        self.special_size = max(1, int(math.floor(3 * delta * capacity / 4)))
        self.primary_size = capacity - self.special_size

        total_buckets = self.primary_size // self.beta
        a1 = (
            total_buckets / (4 * (1 - (0.75) ** self.alpha))
            if self.alpha > 0
            else total_buckets
        )
        self.levels = []
        self.level_bucket_counts = []
        self.level_salts = []
        remaining_buckets = total_buckets
        for i in range(self.alpha):
            a_i = max(1, int(round(a1 * (0.75) ** i)))
            if remaining_buckets <= 0 or a_i <= 0:
                break
            a_i = min(a_i, remaining_buckets)
            self.level_bucket_counts.append(a_i)
            level_size = a_i * self.beta
            level_array = [None] * level_size
            self.levels.append(level_array)
            self.level_salts.append(random.randint(0, sys.maxsize))
            remaining_buckets -= a_i
        if remaining_buckets > 0 and self.levels:
            extra = remaining_buckets * self.beta
            self.levels[-1].extend([None] * extra)
            self.level_bucket_counts[-1] += remaining_buckets
            remaining_buckets = 0

        self.special_array = [None] * self.special_size
        self.special_salt = random.randint(0, sys.maxsize)
        self.special_occupancy = 0

    def _hash_level(self, key, level_index):
        return hash((key, self.level_salts[level_index])) & 0x7FFFFFFF

    def _hash_special(self, key):
        return hash((key, self.special_salt)) & 0x7FFFFFFF

    def __setitem__(self, key, value):
        self.insert(key, value)

    def insert(self, key, value):
        if self.num_inserts >= self.max_inserts:
            raise RuntimeError(
                "Hash table is full (maximum allowed insertions reached)."
            )
        for i in range(len(self.levels)):
            level = self.levels[i]
            num_buckets = self.level_bucket_counts[i]
            bucket_index = self._hash_level(key, i) % num_buckets
            start = bucket_index * self.beta
            end = start + self.beta
            for idx in range(start, end):
                if level[idx] is None:
                    level[idx] = (key, value)
                    self.num_inserts += 1
                    return True
                elif level[idx][0] == key:
                    level[idx] = (key, value)
                    return True
        special = self.special_array
        size = len(special)
        probe_limit = int(max(1, math.ceil(math.log(math.log(self.capacity + 1) + 1))))
        for j in range(probe_limit):
            idx = (self._hash_special(key) + j) % size
            if special[idx] is None:
                special[idx] = (key, value)
                self.special_occupancy += 1
                self.num_inserts += 1
                return True
            elif special[idx][0] == key:
                special[idx] = (key, value)
                return True
        idx1 = self._hash_special(key) % size
        idx2 = (self._hash_special(key) + 1) % size
        if special[idx1] is None:
            special[idx1] = (key, value)
            self.special_occupancy += 1
            self.num_inserts += 1
            return True
        elif special[idx2] is None:
            special[idx2] = (key, value)
            self.special_occupancy += 1
            self.num_inserts += 1
            return True
        else:
            raise RuntimeError("Special array insertion failed; table is full.")

    def __getitem__(self, key):
        ret = self.search(key)
        if ret is None:
            raise KeyError(key)
        return ret

    def get(self, key, default=None):
        return self.search(key) or default

    def search(self, key):
        for i in range(len(self.levels)):
            level = self.levels[i]
            num_buckets = self.level_bucket_counts[i]
            bucket_index = self._hash_level(key, i) % num_buckets
            start = bucket_index * self.beta
            end = start + self.beta
            for idx in range(start, end):
                if level[idx] is None:
                    break
                elif level[idx][0] == key:
                    return level[idx][1]
        special = self.special_array
        size = len(special)
        probe_limit = int(max(1, math.ceil(math.log(math.log(self.capacity + 1) + 1))))
        for j in range(probe_limit):
            idx = (self._hash_special(key) + j) % size
            if special[idx] is None:
                break
            elif special[idx][0] == key:
                return special[idx][1]
        idx1 = self._hash_special(key) % size
        idx2 = (self._hash_special(key) + 1) % size
        if special[idx1] is not None and special[idx1][0] == key:
            return special[idx1][1]
        if special[idx2] is not None and special[idx2][0] == key:
            return special[idx2][1]
        return None

    def __contains__(self, key):
        return self.search(key) is not None

    def __len__(self):
        return self.num_inserts



from setuptools import setup, find_packages

setup(
    name='optopenhash',
    version='0.2.0',
    author='Matthew Stern',
    author_email='matthew.stern@ztsystems.com',
    description=('A Python package implementing improved open‐addressing hash tables '
                 'based on the paper "Optimal Bounds for Open Addressing Without Reordering".'),
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/sternma/optopenhash',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import pytest
from optopenhash import ElasticHashTable, FunnelHashTable


def test_elastic_hash_table():
    capacity = 1000
    delta = 0.1
    et = ElasticHashTable(capacity, delta)
    n_insert = capacity - int(delta * capacity)
    for i in range(n_insert):
        if i % 2:
            et.insert(f"key{i}", f"value{i}")
        else:
            et[f"key{i}"] = f"value{i}"
    # Check that inserted keys are found
    for i in range(n_insert):
        if i % 2:
            assert et[f"key{i}"] == f"value{i}"
        else:
            assert et.search(f"key{i}") == f"value{i}"
    # A key that was not inserted should return None
    assert et.search("nonexistent") is None


def test_funnel_hash_table():
    capacity = 1000
    delta = 0.1
    ft = FunnelHashTable(capacity, delta)
    n_insert = capacity - int(delta * capacity)
    for i in range(n_insert):
        if i % 2:
            ft.insert(f"key{i}", f"value{i}")
        else:
            ft[f"key{i}"] = f"value{i}"
    # Check that inserted keys are found
    for i in range(n_insert):
        if i % 2:
            assert ft[f"key{i}"] == f"value{i}"
        else:
            assert ft.search(f"key{i}") == f"value{i}"
    # A key that was not inserted should return None
    assert ft.search("nonexistent") is None


if __name__ == "__main__":
    pytest.main([__file__])
